from datetime import datetime
from typing import List, Optional, Tuple, cast

from pydantic import UUID4
from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql.sqltypes import JSON, String

from fief.models.base import WorkspaceBase
from fief.models.client import Client
from fief.models.generics import GUID, CreatedUpdatedAt, TIMESTAMPAware, UUIDModel
from fief.models.user import User


class AuthorizationCode(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "authorization_codes"

    code: str = Column(
        String(length=255),
        nullable=False,
        index=True,
        unique=True,
    )
    c_hash: str = Column(String(length=255), nullable=False)
    redirect_uri: str = Column(String(length=2048), nullable=False)
    scope: List[str] = Column(JSON, nullable=False, default=list)
    authenticated_at: datetime = Column(TIMESTAMPAware(timezone=True), nullable=False)
    nonce: Optional[str] = Column(String(length=255), nullable=True)
    code_challenge: Optional[str] = Column(String(length=255), nullable=True)
    code_challenge_method: Optional[str] = Column(String(length=255), nullable=True)

    user_id: UUID4 = Column(GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user: User = relationship("User")

    client_id: UUID4 = Column(GUID, ForeignKey(Client.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    client: Client = relationship("Client", lazy="joined")

    def get_code_challenge_tuple(self) -> Optional[Tuple[str, str]]:
        if self.code_challenge is not None:
            return (self.code_challenge, cast(str, self.code_challenge_method))
        return None



from typing import Dict, Literal

HYBRID_RESPONSE_TYPES = ["code id_token", "code token", "code id_token token"]
ALLOWED_RESPONSE_TYPES = ["code"] + HYBRID_RESPONSE_TYPES
NONCE_REQUIRED_RESPONSE_TYPES = HYBRID_RESPONSE_TYPES

DEFAULT_RESPONSE_MODE: Dict[str, Literal["query", "fragment"]] = {
    "code": "query",
    "code id_token": "fragment",
    "code token": "fragment",
    "code id_token token": "fragment",
}



from fief.tasks.base import SendTask, send_task
from fief.tasks.forgot_password import on_after_forgot_password
from fief.tasks.register import on_after_register

__all__ = [
    "send_task",
    "SendTask",
    "on_after_forgot_password",
    "on_after_register",
]



from typing import List, Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import UserField


class UserFieldManager(BaseManager[UserField], UUIDManagerMixin[UserField]):
    model = UserField

    async def get_by_slug(self, slug: str) -> Optional[UserField]:
        statement = select(UserField).where(UserField.slug == slug)
        return await self.get_one_or_none(statement)

    async def get_registration_fields(self) -> List[UserField]:
        statement = select(UserField)
        user_fields = await self.list(statement)
        return [
            user_field
            for user_field in user_fields
            if user_field.configuration["at_registration"]
        ]

    async def get_update_fields(self) -> List[UserField]:
        statement = select(UserField)
        user_fields = await self.list(statement)
        return [
            user_field
            for user_field in user_fields
            if user_field.configuration["at_update"]
        ]



from datetime import datetime
from typing import Any, Dict, Generic, List, Optional, TypeVar

import phonenumbers
import pycountry
import pytz
from pydantic import UUID4, BaseModel, Field, PydanticValueError
from pydantic.generics import GenericModel

PM = TypeVar("PM", bound=BaseModel)


class UUIDSchema(BaseModel):
    id: UUID4

    class Config:
        orm_mode = True


class CreatedUpdatedAt(BaseModel):
    created_at: datetime
    updated_at: datetime


class PaginatedResults(GenericModel, Generic[PM]):
    count: int
    results: List[PM]


class TrueBooleanError(PydanticValueError):
    code = "boolean.must_be_true"
    msg_template = "value must be true"


def true_bool_validator(cls, v: bool):
    if v is False:
        raise TrueBooleanError()
    return v


class PhoneNumberError(PydanticValueError):
    code = "phone_number.invalid"
    msg_template = "value is not a valid phone number"


class PhoneNumberMissingRegionError(PhoneNumberError):
    code = "phone_number.missing_region"
    msg_template = "value is missing the country code"


class PhoneNumber(str):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        field_schema.update(type="string")

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> str:
        try:
            parsed = phonenumbers.parse(value)
        except phonenumbers.phonenumberutil.NumberParseException as e:
            raise PhoneNumberMissingRegionError() from e
        if not phonenumbers.is_valid_number(parsed):
            raise PhoneNumberError()
        return phonenumbers.format_number(parsed, phonenumbers.PhoneNumberFormat.E164)


class CountryCodeError(PydanticValueError):
    code = "country_code.invalid"
    msg_template = "value is not a valid country code"


class CountryCode(str):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        countries = sorted(pycountry.countries, key=lambda c: c.name)
        field_schema.update(
            type="enum",
            enum=[country.alpha_2 for country in countries],
            countries=[
                {"name": country.name, "alpha_2": country.alpha_2}
                for country in countries
            ],
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> str:
        if pycountry.countries.get(alpha_2=value) is None:
            raise CountryCodeError()
        return value


class Address(BaseModel):
    line1: str = Field(..., min_length=1)
    line2: Optional[str] = Field(None, min_length=1)
    postal_code: str = Field(..., min_length=1)
    city: str = Field(..., min_length=1)
    state: Optional[str] = Field(None, min_length=1)
    country: CountryCode


class TimezoneError(PydanticValueError):
    code = "timezone.invalid"
    msg_template = "value is not a valid timezone"


class Timezone(str):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        field_schema.update(
            type="enum", enum=sorted(pytz.common_timezones), title="timezone"
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> str:
        if not isinstance(value, str):
            raise TypeError()
        try:
            timezone = pytz.timezone(value)
        except pytz.exceptions.UnknownTimeZoneError as e:
            raise TimezoneError() from e
        else:
            return str(timezone)



from fastapi import Depends

from fief.db import AsyncSession
from fief.dependencies.current_workspace import get_current_workspace_session
from fief.managers import (
    AuthorizationCodeManager,
    ClientManager,
    GrantManager,
    LoginSessionManager,
    RefreshTokenManager,
    SessionTokenManager,
    TenantManager,
    UserFieldManager,
    UserManager,
    get_manager,
)


async def get_tenant_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> TenantManager:
    return get_manager(TenantManager, session)


async def get_client_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> ClientManager:
    return get_manager(ClientManager, session)


async def get_user_field_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> UserFieldManager:
    return get_manager(UserFieldManager, session)


async def get_user_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> UserManager:
    return get_manager(UserManager, session)


async def get_login_session_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> LoginSessionManager:
    return get_manager(LoginSessionManager, session)


async def get_authorization_code_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> AuthorizationCodeManager:
    return get_manager(AuthorizationCodeManager, session)


async def get_refresh_token_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> RefreshTokenManager:
    return get_manager(RefreshTokenManager, session)


async def get_session_token_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> SessionTokenManager:
    return get_manager(SessionTokenManager, session)


async def get_grant_manager(
    session: AsyncSession = Depends(get_current_workspace_session),
) -> GrantManager:
    return get_manager(GrantManager, session)



from datetime import datetime
from typing import List

from pydantic import UUID4
from sqlalchemy import JSON, Column, ForeignKey, String
from sqlalchemy.orm import relationship

from fief.models.base import WorkspaceBase
from fief.models.client import Client
from fief.models.generics import GUID, CreatedUpdatedAt, TIMESTAMPAware, UUIDModel
from fief.models.user import User


class RefreshToken(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "refresh_tokens"

    token: str = Column(
        String(length=255),
        nullable=False,
        index=True,
        unique=True,
    )
    expires_at: datetime = Column(
        TIMESTAMPAware(timezone=True), nullable=False, index=True
    )
    scope: List[str] = Column(JSON, nullable=False, default=list)
    authenticated_at: datetime = Column(TIMESTAMPAware(timezone=True), nullable=False)

    user_id: UUID4 = Column(GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user: User = relationship("User")

    client_id: UUID4 = Column(GUID, ForeignKey(Client.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    client: Client = relationship("Client", lazy="joined")



from typing import List, Tuple

from fastapi import Depends, HTTPException, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    Ordering,
    Pagination,
    get_ordering,
    get_paginated_objects,
    get_pagination,
)
from fief.dependencies.workspace_managers import get_client_manager
from fief.managers import ClientManager
from fief.models import Client


async def get_paginated_clients(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(get_ordering),
    manager: ClientManager = Depends(get_client_manager),
) -> Tuple[List[Client], int]:
    statement = select(Client)
    return await get_paginated_objects(statement, pagination, ordering, manager)


async def get_client_by_id_or_404(
    id: UUID4,
    manager: ClientManager = Depends(get_client_manager),
) -> Client:
    client = await manager.get_by_id(id)

    if client is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return client



import uuid

import dramatiq

from fief.tasks.base import TaskBase


class OnAfterForgotPasswordTask(TaskBase):
    __name__ = "on_after_forgot_password"

    async def run(self, user_id: str, workspace_id: str, reset_url: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        user = await self._get_user(uuid.UUID(user_id), workspace)
        tenant = await self._get_tenant(user.tenant_id, workspace)

        translations = user.get_preferred_translations()
        title = translations.gettext("Reset your %(tenant)s's password") % {
            "tenant": tenant.name
        }
        context = {"tenant": tenant, "title": title, "reset_url": reset_url}
        html = self._render_email_template(
            "forgot_password.html", translations, context
        )
        self.email_provider.send_email(
            sender=("contact@fief.dev", None),
            recipient=(user.email, None),
            subject=title,
            html=html,
        )


on_after_forgot_password = dramatiq.actor(OnAfterForgotPasswordTask())



from datetime import datetime, timedelta, timezone

from fastapi import APIRouter, Depends, Response

from fief.crypto.access_token import generate_access_token
from fief.crypto.id_token import generate_id_token
from fief.crypto.token import generate_token
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.token import (
    GrantRequest,
    get_user_from_grant_request,
    validate_grant_request,
)
from fief.dependencies.workspace_managers import get_refresh_token_manager
from fief.managers import RefreshTokenManager
from fief.models import RefreshToken, Tenant, User, Workspace
from fief.schemas.auth import TokenResponse
from fief.settings import settings

router = APIRouter()


@router.post("/token", name="auth:token")
async def token(
    response: Response,
    grant_request: GrantRequest = Depends(validate_grant_request),
    user: User = Depends(get_user_from_grant_request),
    refresh_token_manager: RefreshTokenManager = Depends(get_refresh_token_manager),
    workspace: Workspace = Depends(get_current_workspace),
    tenant: Tenant = Depends(get_current_tenant),
):
    scope = grant_request["scope"]
    authenticated_at = grant_request["authenticated_at"]
    nonce = grant_request["nonce"]
    c_hash = grant_request["c_hash"]
    client = grant_request["client"]

    tenant_host = tenant.get_host(workspace.domain)
    access_token = generate_access_token(
        tenant.get_sign_jwk(),
        tenant_host,
        client,
        user,
        scope,
        settings.access_id_token_lifetime_seconds,
    )
    id_token = generate_id_token(
        tenant.get_sign_jwk(),
        tenant_host,
        client,
        authenticated_at,
        user,
        settings.access_id_token_lifetime_seconds,
        nonce=nonce,
        c_hash=c_hash,
        access_token=access_token,
        encryption_key=client.get_encrypt_jwk(),
    )
    token_response = TokenResponse(
        access_token=access_token,
        id_token=id_token,
        expires_in=settings.access_id_token_lifetime_seconds,
    )

    if "offline_access" in scope:
        expires_at = datetime.now(timezone.utc) + timedelta(
            seconds=settings.refresh_token_lifetime_seconds
        )
        token, token_hash = generate_token()
        refresh_token = RefreshToken(
            token=token_hash,
            expires_at=expires_at,
            scope=scope,
            user_id=user.id,
            client_id=client.id,
            authenticated_at=authenticated_at,
        )
        refresh_token = await refresh_token_manager.create(refresh_token)
        token_response.refresh_token = token

    response.headers["Cache-Control"] = "no-store"
    response.headers["Pragma"] = "no-cache"
    return token_response.dict(exclude_none=True)



from typing import List, Optional, Tuple

from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse
from fastapi.security.oauth2 import OAuth2PasswordRequestForm
from pydantic import AnyUrl

from fief.apps.auth.templates import templates
from fief.csrf import check_csrf
from fief.dependencies.auth import (
    check_unsupported_request_parameter,
    get_authorize_client,
    get_authorize_code_challenge,
    get_authorize_prompt,
    get_authorize_redirect_uri,
    get_authorize_response_mode,
    get_authorize_response_type,
    get_authorize_scope,
    get_authorize_screen,
    get_consent_action,
    get_consent_prompt,
    get_login_session,
    get_needs_consent,
    get_nonce,
    has_valid_session_token,
)
from fief.dependencies.authentication_flow import get_authentication_flow
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.locale import get_gettext, get_translations
from fief.dependencies.session_token import get_session_token
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.users import UserManager, get_user_manager
from fief.dependencies.workspace_managers import get_session_token_manager
from fief.exceptions import LoginException, LogoutException
from fief.locale import Translations
from fief.managers.session_token import SessionTokenManager
from fief.models import Client, LoginSession, Tenant, Workspace
from fief.models.session_token import SessionToken
from fief.schemas.auth import LoginError, LogoutError
from fief.services.authentication_flow import AuthenticationFlow
from fief.settings import settings

router = APIRouter(dependencies=[Depends(check_csrf), Depends(get_translations)])


@router.get(
    "/authorize",
    name="auth:authorize",
    dependencies=[Depends(check_unsupported_request_parameter)],
)
async def authorize(
    request: Request,
    response_type: str = Depends(get_authorize_response_type),
    client: Client = Depends(get_authorize_client),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    scope: List[str] = Depends(get_authorize_scope),
    prompt: Optional[str] = Depends(get_authorize_prompt),
    screen: str = Depends(get_authorize_screen),
    code_challenge_tuple: Optional[Tuple[str, str]] = Depends(
        get_authorize_code_challenge
    ),
    nonce: Optional[str] = Depends(get_nonce),
    state: Optional[str] = Query(None),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    has_valid_session_token: bool = Depends(has_valid_session_token),
):
    tenant = client.tenant

    if has_valid_session_token and prompt != "login":
        redirection = tenant.url_for(request, "auth:consent.get")
    elif screen == "register":
        redirection = tenant.url_for(request, "register:get")
    else:
        redirection = tenant.url_for(request, "auth:login.get")

    response = RedirectResponse(url=redirection, status_code=status.HTTP_302_FOUND)
    response = await authentication_flow.create_login_session(
        response,
        response_type=response_type,
        response_mode=response_mode,
        redirect_uri=redirect_uri,
        scope=scope,
        state=state,
        nonce=nonce,
        code_challenge_tuple=code_challenge_tuple,
        client=client,
    )

    return response


@router.get("/login", name="auth:login.get")
async def get_login(
    request: Request,
    login_session: LoginSession = Depends(get_login_session),
    translations: Translations = Depends(get_translations),
):
    return templates.LocaleTemplateResponse(
        "login.html",
        {"request": request, "tenant": login_session.client.tenant},
        translations=translations,
    )


@router.post("/login", name="auth:login.post")
async def post_login(
    request: Request,
    credentials: OAuth2PasswordRequestForm = Depends(OAuth2PasswordRequestForm),
    login_session: LoginSession = Depends(get_login_session),
    user_manager: UserManager = Depends(get_user_manager),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    session_token: Optional[SessionToken] = Depends(get_session_token),
    tenant: Tenant = Depends(get_current_tenant),
    _=Depends(get_gettext),
):
    user = await user_manager.authenticate(credentials)

    if user is None or not user.is_active:
        raise LoginException(
            LoginError.get_bad_credentials(_("Invalid email or password")),
            login_session.client.tenant,
        )
    # if requires_verification and not user.is_verified:
    #     raise HTTPException(
    #         status_code=status.HTTP_400_BAD_REQUEST,
    #         detail=ErrorCode.LOGIN_USER_NOT_VERIFIED,
    #     )

    response = RedirectResponse(
        tenant.url_for(request, "auth:consent.get"),
        status_code=status.HTTP_302_FOUND,
    )
    response = await authentication_flow.rotate_session_token(
        response, user.id, session_token=session_token
    )

    return response


@router.get("/consent", name="auth:consent.get")
async def get_consent(
    request: Request,
    login_session: LoginSession = Depends(get_login_session),
    session_token: Optional[SessionToken] = Depends(get_session_token),
    prompt: Optional[str] = Depends(get_consent_prompt),
    needs_consent: bool = Depends(get_needs_consent),
    tenant: Tenant = Depends(get_current_tenant),
    workspace: Workspace = Depends(get_current_workspace),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    translations: Translations = Depends(get_translations),
):
    if session_token is None:
        return RedirectResponse(
            tenant.url_for(request, "auth:login.get"), status_code=status.HTTP_302_FOUND
        )

    if not needs_consent and prompt != "consent":
        response = await authentication_flow.get_authorization_code_success_redirect(
            login_session=login_session,
            authenticated_at=session_token.created_at,
            user=session_token.user,
            client=login_session.client,
            tenant=tenant,
            workspace=workspace,
        )
        response = await authentication_flow.delete_login_session(
            response, login_session
        )
        return response

    return templates.LocaleTemplateResponse(
        "consent.html",
        {
            "request": request,
            "tenant": login_session.client.tenant,
            "client": login_session.client,
            "scopes": login_session.scope,
        },
        translations=translations,
    )


@router.post("/consent", name="auth:consent.post")
async def post_consent(
    request: Request,
    action: str = Depends(get_consent_action),
    login_session: LoginSession = Depends(get_login_session),
    session_token: Optional[SessionToken] = Depends(get_session_token),
    tenant: Tenant = Depends(get_current_tenant),
    workspace: Workspace = Depends(get_current_workspace),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    _=Depends(get_gettext),
):
    if session_token is None:
        return RedirectResponse(
            tenant.url_for(request, "auth:login.get"), status_code=status.HTTP_302_FOUND
        )

    if action == "allow":
        user_id = session_token.user_id
        await authentication_flow.create_or_update_grant(
            user_id, login_session.client, login_session.scope
        )
        response = await authentication_flow.get_authorization_code_success_redirect(
            login_session=login_session,
            authenticated_at=session_token.created_at,
            user=session_token.user,
            client=login_session.client,
            tenant=tenant,
            workspace=workspace,
        )
    elif action == "deny":
        response = AuthenticationFlow.get_authorization_code_error_redirect(
            login_session.redirect_uri,
            login_session.response_mode,
            "access_denied",
            error_description=_("The user denied access to their data."),
            state=login_session.state,
        )

    response = await authentication_flow.delete_login_session(response, login_session)

    return response


@router.get("/logout", name="auth:logout")
async def logout(
    redirect_uri: Optional[AnyUrl] = Query(None),
    session_token: Optional[SessionToken] = Depends(get_session_token),
    sesstion_token_manager: SessionTokenManager = Depends(get_session_token_manager),
    tenant: Tenant = Depends(get_current_tenant),
    _=Depends(get_gettext),
):
    if redirect_uri is None:
        raise LogoutException(
            LogoutError.get_invalid_request(_("redirect_uri is missing")),
            tenant,
        )

    if session_token is not None:
        await sesstion_token_manager.delete(session_token)

    response = RedirectResponse(redirect_uri, status_code=status.HTTP_302_FOUND)

    response.delete_cookie(
        settings.session_cookie_name,
        domain=settings.session_cookie_domain,
        secure=settings.session_cookie_secure,
        httponly=True,
    )

    return response



"""Add alembic_revision column to Workspace

Revision ID: 696e383aabf2
Revises: 274bc650f9a2
Create Date: 2022-04-20 09:44:10.693572

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "696e383aabf2"
down_revision = "274bc650f9a2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "workspaces",
        sa.Column("alembic_revision", sa.String(length=255), nullable=True),
    )
    op.create_index(
        op.f("ix_workspaces_alembic_revision"),
        "workspaces",
        ["alembic_revision"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_workspaces_alembic_revision"), table_name="workspaces")
    op.drop_column("workspaces", "alembic_revision")
    # ### end Alembic commands ###



from enum import Enum
from pathlib import Path
from typing import Dict, Optional

from sqlalchemy import engine


class DatabaseType(str, Enum):
    POSTGRESQL = "POSTGRESQL"
    MYSQL = "MYSQL"
    SQLITE = "SQLITE"


SYNC_DRIVERS: Dict[DatabaseType, str] = {
    DatabaseType.POSTGRESQL: "postgresql",
    DatabaseType.MYSQL: "mysql+pymysql",
    DatabaseType.SQLITE: "sqlite",
}

ASYNC_DRIVERS: Dict[DatabaseType, str] = {
    DatabaseType.POSTGRESQL: "postgresql+asyncpg",
    DatabaseType.MYSQL: "mysql+aiomysql",
    DatabaseType.SQLITE: "sqlite+aiosqlite",
}


def get_driver(type: DatabaseType, *, asyncio: bool) -> str:
    drivers = ASYNC_DRIVERS if asyncio else SYNC_DRIVERS
    return drivers[type]


def create_database_url(
    type: DatabaseType,
    *,
    asyncio: bool,
    username: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    database: Optional[str] = None,
    path: Optional[Path] = None,
    schema: Optional[str] = None,
) -> engine.URL:
    url = engine.URL.create(
        drivername=get_driver(type, asyncio=asyncio),
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
    )

    dialect_name = url.get_dialect().name
    if dialect_name == "sqlite":
        name = schema if schema is not None else database
        assert name is not None
        assert path is not None
        url = url.set(database=str(path / name))

    return url



import hashlib
import hmac
import secrets
from typing import Tuple

from fief.settings import settings


def get_token_hash(
    token: str, *, secret: str = settings.secret.get_secret_value()
) -> str:
    hash = hmac.new(secret.encode("ascii"), token.encode("ascii"), hashlib.sha256)
    return hash.hexdigest()


def generate_token(
    *, secret: str = settings.secret.get_secret_value()
) -> Tuple[str, str]:
    """
    Generate a token suitable for sensitive values
    like authorization codes or refresh tokens.

    Returns both the actual value and its HMAC-SHA256 hash.
    Only the latter shall be stored in database.
    """
    token = secrets.token_urlsafe()
    return token, get_token_hash(token, secret=secret)



from fastapi import APIRouter, Depends

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.tenant import get_paginated_tenants
from fief.models import Tenant
from fief.schemas.generics import PaginatedResults

router = APIRouter(dependencies=[Depends(is_authenticated_admin)])


@router.get("/", response_model=PaginatedResults[schemas.tenant.Tenant])
async def list_tenants(
    paginated_tenants: PaginatedObjects[Tenant] = Depends(get_paginated_tenants),
) -> PaginatedResults[schemas.tenant.Tenant]:
    tenants, count = paginated_tenants
    return PaginatedResults(
        count=count,
        results=[schemas.tenant.Tenant.from_orm(tenant) for tenant in tenants],
    )



from typing import Optional

from pydantic import UUID4
from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import Grant


class GrantManager(BaseManager[Grant], UUIDManagerMixin[Grant]):
    model = Grant

    async def get_by_user_and_client(
        self, user_id: UUID4, client_id: UUID4
    ) -> Optional[Grant]:
        statement = select(Grant).where(
            Grant.user_id == user_id, Grant.client_id == client_id
        )
        return await self.get_one_or_none(statement)



from fief.managers.admin_api_key import AdminAPIKeyManager
from fief.managers.admin_session_token import AdminSessionTokenManager
from fief.managers.authorization_code import AuthorizationCodeManager
from fief.managers.base import get_manager
from fief.managers.client import ClientManager
from fief.managers.grant import GrantManager
from fief.managers.login_session import LoginSessionManager
from fief.managers.refresh_token import RefreshTokenManager
from fief.managers.session_token import SessionTokenManager
from fief.managers.tenant import TenantManager
from fief.managers.user import UserManager
from fief.managers.user_field import UserFieldManager
from fief.managers.workspace import WorkspaceManager
from fief.managers.workspace_user import WorkspaceUserManager

__all__ = [
    "WorkspaceManager",
    "WorkspaceUserManager",
    "AdminAPIKeyManager",
    "AdminSessionTokenManager",
    "AuthorizationCodeManager",
    "ClientManager",
    "GrantManager",
    "LoginSessionManager",
    "RefreshTokenManager",
    "SessionTokenManager",
    "TenantManager",
    "UserManager",
    "UserFieldManager",
    "get_manager",
]



from fief.models.admin_api_key import AdminAPIKey
from fief.models.admin_session_token import AdminSessionToken
from fief.models.authorization_code import AuthorizationCode
from fief.models.base import MainBase, WorkspaceBase
from fief.models.client import Client, ClientType
from fief.models.generics import M_UUID, M
from fief.models.grant import Grant
from fief.models.login_session import LoginSession
from fief.models.refresh_token import RefreshToken
from fief.models.session_token import SessionToken
from fief.models.tenant import Tenant
from fief.models.user import User
from fief.models.user_field import UserField, UserFieldType
from fief.models.user_field_value import UserFieldValue
from fief.models.workspace import Workspace
from fief.models.workspace_user import WorkspaceUser

__all__ = [
    "Workspace",
    "WorkspaceBase",
    "WorkspaceUser",
    "AdminAPIKey",
    "AdminSessionToken",
    "AuthorizationCode",
    "Client",
    "ClientType",
    "Grant",
    "LoginSession",
    "MainBase",
    "RefreshToken",
    "SessionToken",
    "M",
    "M_UUID",
    "Tenant",
    "User",
    "UserField",
    "UserFieldType",
    "UserFieldValue",
]



"""Add UserFieldValue model

Revision ID: ac76d4c912eb
Revises: b530ce0d8bd9
Create Date: 2022-04-28 15:36:49.761231

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "ac76d4c912eb"
down_revision = "b530ce0d8bd9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_user_field_values",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("value_string", sa.Text(), nullable=True),
        sa.Column("value_integer", sa.Integer(), nullable=True),
        sa.Column("value_boolean", sa.Boolean(), nullable=True),
        sa.Column("value_date", sa.Date(), nullable=True),
        sa.Column(
            "value_datetime",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=True,
        ),
        sa.Column("value_json", sa.JSON(), nullable=True),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("user_field_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_field_id"], ["fief_user_fields.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_field_id", "user_id"),
    )
    op.create_index(
        op.f("ix_fief_user_field_values_created_at"),
        "fief_user_field_values",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_user_field_values_updated_at"),
        "fief_user_field_values",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_fief_user_field_values_updated_at"),
        table_name="fief_user_field_values",
    )
    op.drop_index(
        op.f("ix_fief_user_field_values_created_at"),
        table_name="fief_user_field_values",
    )
    op.drop_table("fief_user_field_values")
    # ### end Alembic commands ###



from typing import Callable, Dict, Type

from fastapi import Request, status
from fastapi.responses import JSONResponse

from fief.apps.auth.templates import templates
from fief.exceptions import (
    AuthorizeException,
    AuthorizeRedirectException,
    ConsentException,
    FormValidationError,
    LoginException,
    LogoutException,
    RegisterException,
    ResetPasswordException,
    TokenRequestException,
)
from fief.services.authentication_flow import AuthenticationFlow

exception_handlers: Dict[Type[Exception], Callable] = {}


async def form_validation_error_handler(request: Request, exc: FormValidationError):
    return templates.LocaleTemplateResponse(
        exc.template,
        {
            "request": request,
            "form_errors": exc.form_errors(),
            "fatal_error": False,
            **exc.context,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
    )


exception_handlers[FormValidationError] = form_validation_error_handler


async def register_exception_handler(request: Request, exc: RegisterException):
    return templates.LocaleTemplateResponse(
        "register.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "form_data": exc.form_data,
            "fatal_error": exc.fatal,
            **exc.context,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[RegisterException] = register_exception_handler


async def authorize_exception_handler(request: Request, exc: AuthorizeException):
    return templates.LocaleTemplateResponse(
        "authorize.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "tenant": exc.tenant,
            "fatal_error": True,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[AuthorizeException] = authorize_exception_handler


async def authorize_redirect_exception_handler(
    request: Request, exc: AuthorizeRedirectException
):
    return AuthenticationFlow.get_authorization_code_error_redirect(
        exc.redirect_uri,
        exc.response_mode,
        exc.error.error,
        error_description=exc.error.error_description,
        error_uri=exc.error.error_uri,
        state=exc.state,
    )


exception_handlers[AuthorizeRedirectException] = authorize_redirect_exception_handler


async def login_exception_handler(request: Request, exc: LoginException):
    return templates.LocaleTemplateResponse(
        "login.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "tenant": exc.tenant,
            "fatal_error": exc.fatal,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[LoginException] = login_exception_handler


async def consent_exception_handler(request: Request, exc: ConsentException):
    return templates.LocaleTemplateResponse(
        "consent.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "client": exc.client,
            "scopes": exc.scope,
            "tenant": exc.tenant,
            "fatal_error": exc.fatal,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[ConsentException] = consent_exception_handler


async def token_request_exception_handler(request: Request, exc: TokenRequestException):
    return JSONResponse(
        status_code=status.HTTP_400_BAD_REQUEST,
        content=exc.error.dict(exclude_none=True),
    )


exception_handlers[TokenRequestException] = token_request_exception_handler


async def reset_password_exception_handler(
    request: Request, exc: ResetPasswordException
):
    return templates.LocaleTemplateResponse(
        "reset_password.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "form_data": exc.form_data,
            "tenant": exc.tenant,
            "fatal_error": exc.fatal,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[ResetPasswordException] = reset_password_exception_handler


async def logout_exception_handler(request: Request, exc: LogoutException):
    return templates.LocaleTemplateResponse(
        "logout.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "tenant": exc.tenant,
            "fatal_error": True,
        },
        translations=request.scope["translations"],
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[LogoutException] = logout_exception_handler

__all__ = ["exception_handlers"]



from fief.db.main import main_async_session_maker
from fief.db.workspace import get_workspace_session
from fief.dependencies.users import get_user_db, get_user_manager
from fief.locale import get_preferred_translations
from fief.managers import (
    ClientManager,
    TenantManager,
    WorkspaceManager,
    WorkspaceUserManager,
)
from fief.models import Client, User, Workspace, WorkspaceUser
from fief.schemas.user import UserCreateInternal
from fief.schemas.workspace import WorkspaceCreate
from fief.services.workspace_db import WorkspaceDatabase
from fief.settings import settings
from fief.tasks import send_task


class MainWorkspaceError(Exception):
    pass


class MainWorkspaceAlreadyExists(MainWorkspaceError):
    pass


class MainWorkspaceDoesNotExist(MainWorkspaceError):
    pass


class MainWorkspaceDoesNotHaveDefaultTenant(MainWorkspaceError):
    pass


class MainWorkspaceClientDoesNotExist(MainWorkspaceError):
    pass


class CreateMainFiefUserError(MainWorkspaceError):
    pass


async def get_main_fief_workspace() -> Workspace:
    async with main_async_session_maker() as session:
        workspace_manager = WorkspaceManager(session)
        workspace = await workspace_manager.get_main()

        if workspace is None:
            raise MainWorkspaceDoesNotExist()

        return workspace


async def get_main_fief_client() -> Client:
    workspace = await get_main_fief_workspace()
    async with get_workspace_session(workspace) as session:
        client_manager = ClientManager(session)
        client = await client_manager.get_by_client_id(settings.fief_client_id)

        if client is None:
            raise MainWorkspaceClientDoesNotExist()

        return client


async def create_main_fief_workspace() -> Workspace:
    from fief.services.workspace_creation import WorkspaceCreation

    async with main_async_session_maker() as session:
        workspace_manager = WorkspaceManager(session)
        workspace_user_manager = WorkspaceUserManager(session)
        workspace = await workspace_manager.get_main()

        if workspace is not None:
            raise MainWorkspaceAlreadyExists()

        workspace_create = WorkspaceCreate(name="Fief")
        workspace_db = WorkspaceDatabase()
        workspace_creation = WorkspaceCreation(
            workspace_manager, workspace_user_manager, workspace_db
        )

        workspace = await workspace_creation.create(
            workspace_create,
            default_domain=settings.fief_domain,
            default_client_id=settings.fief_client_id,
            default_client_secret=settings.fief_client_secret,
            default_encryption_key=settings.fief_encryption_key,
            default_redirect_uris=[
                f"http://{settings.fief_domain}/admin/api/auth/callback",
                f"https://{settings.fief_domain}/admin/api/auth/callback",
                "http://localhost:8000/docs/oauth2-redirect",
            ],
        )

    return workspace


async def create_main_fief_user(email: str, password: str) -> User:
    workspace = await get_main_fief_workspace()

    async with main_async_session_maker() as session:
        workspace_user_manager = WorkspaceUserManager(session)

        async with get_workspace_session(workspace) as session:
            tenant_manager = TenantManager(session)
            tenant = await tenant_manager.get_default()

            if tenant is None:
                raise MainWorkspaceDoesNotHaveDefaultTenant()

            user_db = await get_user_db(session, tenant)
            user_manager = await get_user_manager(
                user_db,
                tenant,
                workspace,
                get_preferred_translations(["en"]),
                send_task,
            )
            user = await user_manager.create(
                UserCreateInternal(email=email, password=password, tenant_id=tenant.id)
            )

        workspace_user = WorkspaceUser(workspace_id=workspace.id, user_id=user.id)
        await workspace_user_manager.create(workspace_user)

    return user



import sentry_sdk
from fastapi import FastAPI
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.integrations.redis import RedisIntegration

from fief import __version__
from fief.apps import admin_app, admin_frontend_app, auth_app
from fief.db.workspace import workspace_engine_manager
from fief.settings import settings

sentry_sdk.init(
    dsn=settings.sentry_dsn_server,
    environment=settings.environment.value,
    release=__version__,
    integrations=[RedisIntegration()],
)

app = FastAPI(openapi_url=None)

app.add_middleware(SentryAsgiMiddleware)

app.mount("/admin/api", admin_app)
app.mount("/admin", admin_frontend_app)
app.mount("/", auth_app)


@app.on_event("shutdown")
async def on_shutdown():
    await workspace_engine_manager.close_all()


__all__ = ["app"]



from typing import Optional, Protocol, Tuple


class EmailProvider(Protocol):
    def send_email(
        self,
        *,
        sender: Tuple[str, Optional[str]],
        recipient: Tuple[str, Optional[str]],
        subject: str,
        html: Optional[str] = None,
        text: Optional[str] = None,
    ):
        ...


class EmailError(Exception):
    def __init__(self, message: str):
        self.message = message


class SendEmailError(EmailError):
    pass



from typing import Optional, Tuple

from alembic import command
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy import create_engine, exc, inspect
from sqlalchemy.engine import URL, Engine
from sqlalchemy.schema import CreateSchema

from fief.paths import ALEMBIC_CONFIG_FILE


class WorkspaceDatabaseError(Exception):
    pass


class WorkspaceDatabaseConnectionError(WorkspaceDatabaseError):
    def __init__(self, message: str) -> None:
        self.message = message


class WorkspaceDatabase:
    def migrate(self, database_url: URL, schema_name: str) -> str:
        try:
            engine = self.get_engine(database_url, schema_name)
            with engine.begin() as connection:
                config = self._get_alembic_base_config()
                config.attributes["configure_logger"] = False
                config.attributes["connection"] = connection
                command.upgrade(config, "head")
        except exc.OperationalError as e:
            raise WorkspaceDatabaseConnectionError(str(e)) from e

        return self.get_latest_revision()

    def get_engine(self, database_url: URL, schema_name: str) -> Engine:
        self._ensure_schema(database_url, schema_name)

        connect_args = {}

        dialect_name = database_url.get_dialect().name
        if dialect_name == "postgresql":
            connect_args["options"] = f"-csearch_path={schema_name}"
        elif dialect_name == "mysql":
            database_url = database_url.set(database=schema_name)

        return create_engine(database_url, connect_args=connect_args)

    def check_connection(self, database_url: URL) -> Tuple[bool, Optional[str]]:
        try:
            engine = create_engine(database_url)
            with engine.begin():
                return True, None
        except exc.OperationalError as e:
            return False, str(e)

    def get_latest_revision(self) -> str:
        config = self._get_alembic_base_config()
        script = ScriptDirectory.from_config(config)
        with EnvironmentContext(config, script) as environment:
            return str(environment.get_head_revision())

    def _get_alembic_base_config(self) -> Config:
        return Config(ALEMBIC_CONFIG_FILE, ini_section="workspace")

    def _ensure_schema(self, database_url: URL, schema_name: str):
        engine = create_engine(database_url, connect_args={"connect_timeout": 5})

        dialect_name = database_url.get_dialect().name
        if dialect_name == "sqlite":
            return

        inspector = inspect(engine)
        schemas = inspector.get_schema_names()
        if schema_name not in schemas:
            with engine.begin() as connection:
                connection.execute(CreateSchema(schema_name))



from fastapi_users.password import PasswordHelper
from passlib.context import CryptContext

context = CryptContext(schemes=["argon2", "bcrypt"], deprecated="auto")
password_helper = PasswordHelper(context)



from typing import Optional

from fastapi_users.router import ErrorCode
from pydantic import BaseModel


class RegisterError(BaseModel):
    error: ErrorCode
    error_description: Optional[str] = None
    error_uri: Optional[str] = None

    @classmethod
    def get_user_already_exists(cls, error_description: Optional[str] = None):
        return cls(
            error=ErrorCode.REGISTER_USER_ALREADY_EXISTS,
            error_description=error_description,
        )

    @classmethod
    def get_invalid_password(cls, error_description: Optional[str] = None):
        return cls(
            error=ErrorCode.REGISTER_INVALID_PASSWORD,
            error_description=error_description,
        )



"""Initial migration

Revision ID: 274bc650f9a2
Revises:
Create Date: 2022-03-17 14:41:26.913781

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

import fief

# revision identifiers, used by Alembic.
revision = "274bc650f9a2"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    if connection.dialect.name == "postgresql":
        database_type_enum = postgresql.ENUM(
            "POSTGRESQL", "MYSQL", "SQLITE", name="databasetype", create_type=False
        )
        database_type_enum.create(connection, checkfirst=True)
    else:
        database_type_enum = sa.Enum(
            "POSTGRESQL", "MYSQL", "SQLITE", name="databasetype"
        )

    op.create_table(
        "admin_session_tokens",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column("raw_tokens", sa.Text(), nullable=False),
        sa.Column("raw_userinfo", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("token"),
    )
    op.create_index(
        op.f("ix_admin_session_tokens_created_at"),
        "admin_session_tokens",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_admin_session_tokens_updated_at"),
        "admin_session_tokens",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "workspaces",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("domain", sa.String(length=255), nullable=False),
        sa.Column("database_type", database_type_enum, nullable=True),
        sa.Column("database_host", sa.Text(), nullable=True),
        sa.Column("database_port", sa.Text(), nullable=True),
        sa.Column("database_username", sa.Text(), nullable=True),
        sa.Column("database_password", sa.Text(), nullable=True),
        sa.Column("database_name", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_workspaces_created_at"), "workspaces", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_workspaces_updated_at"), "workspaces", ["updated_at"], unique=False
    )
    op.create_table(
        "admin_api_key",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column("workspace_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["workspace_id"], ["workspaces.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("token"),
    )
    op.create_index(
        op.f("ix_admin_api_key_created_at"),
        "admin_api_key",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_admin_api_key_updated_at"),
        "admin_api_key",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "workspace_users",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("workspace_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["workspace_id"], ["workspaces.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_workspace_users_created_at"),
        "workspace_users",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_workspace_users_updated_at"),
        "workspace_users",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_workspace_users_updated_at"), table_name="workspace_users")
    op.drop_index(op.f("ix_workspace_users_created_at"), table_name="workspace_users")
    op.drop_table("workspace_users")
    op.drop_index(op.f("ix_admin_api_key_updated_at"), table_name="admin_api_key")
    op.drop_index(op.f("ix_admin_api_key_created_at"), table_name="admin_api_key")
    op.drop_table("admin_api_key")
    op.drop_index(op.f("ix_workspaces_updated_at"), table_name="workspaces")
    op.drop_index(op.f("ix_workspaces_created_at"), table_name="workspaces")
    op.drop_table("workspaces")
    op.drop_index(
        op.f("ix_admin_session_tokens_updated_at"), table_name="admin_session_tokens"
    )
    op.drop_index(
        op.f("ix_admin_session_tokens_created_at"), table_name="admin_session_tokens"
    )
    op.drop_table("admin_session_tokens")

    connection = op.get_bind()
    if connection.dialect.name == "postgresql":
        database_type_enum = postgresql.ENUM(
            "POSTGRESQL", "MYSQL", "SQLITE", name="databasetype", create_type=False
        )
        database_type_enum.drop(op.get_bind(), checkfirst=True)
    # ### end Alembic commands ###



from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import AdminAPIKey


class AdminAPIKeyManager(BaseManager[AdminAPIKey], UUIDManagerMixin[AdminAPIKey]):
    model = AdminAPIKey

    async def get_by_token(self, token: str) -> Optional[AdminAPIKey]:
        statement = select(AdminAPIKey).where(AdminAPIKey.token == token)
        return await self.get_one_or_none(statement)



from typing import Any, List, Literal, Optional, Tuple, Type

from fastapi import Depends, HTTPException, status
from fastapi.exceptions import RequestValidationError
from pydantic import UUID4, ValidationError, create_model, validator
from sqlalchemy import select

from fief.dependencies.pagination import (
    Ordering,
    Pagination,
    get_ordering,
    get_paginated_objects,
    get_pagination,
)
from fief.dependencies.workspace_managers import get_user_field_manager
from fief.managers import UserFieldManager
from fief.models import UserField
from fief.models.user_field import UserFieldType
from fief.schemas.generics import true_bool_validator
from fief.schemas.user import UF, UserCreate, UserCreateInternal, UserFields, UserUpdate
from fief.schemas.user_field import (
    USER_FIELD_CAN_HAVE_DEFAULT,
    USER_FIELD_TYPE_MAP,
    UserFieldConfigurationBase,
    UserFieldConfigurationChoice,
    UserFieldConfigurationDefault,
    UserFieldCreate,
    UserFieldUpdate,
    get_user_field_pydantic_type,
)


async def get_paginated_user_fields(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(get_ordering),
    manager: UserFieldManager = Depends(get_user_field_manager),
) -> Tuple[List[UserField], int]:
    statement = select(UserField)
    return await get_paginated_objects(statement, pagination, ordering, manager)


async def get_user_field_by_id_or_404(
    id: UUID4,
    manager: UserFieldManager = Depends(get_user_field_manager),
) -> UserField:
    user_field = await manager.get_by_id(id)

    if user_field is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return user_field


async def get_user_fields(
    manager: UserFieldManager = Depends(get_user_field_manager),
) -> List[UserField]:
    return await manager.all()


async def get_user_field_create_internal_model(user_field_create: UserFieldCreate):
    user_field_type = user_field_create.type
    configuration_type = UserFieldConfigurationBase
    if user_field_type == UserFieldType.CHOICE:
        configuration_type = UserFieldConfigurationChoice
    elif USER_FIELD_CAN_HAVE_DEFAULT[user_field_type]:
        configuration_type = UserFieldConfigurationDefault[
            USER_FIELD_TYPE_MAP[user_field_type]  # type: ignore
        ]

    return create_model(
        "UserFieldCreateInternal",
        type=(Literal[user_field_type], ...),  # type: ignore
        configuration=(configuration_type, ...),
        __base__=UserFieldCreate,
    )


async def get_validated_user_field_create(
    user_field_create: UserFieldCreate,
    user_field_create_internal_model: Type[UserFieldCreate] = Depends(
        get_user_field_create_internal_model
    ),
) -> UserFieldCreate:
    body_model = create_model(
        "UserFieldCreateInternalBody",
        body=(user_field_create_internal_model, ...),
    )
    try:
        validated_user_field_create = body_model(body=user_field_create.dict())
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_field_create.body  # type: ignore


async def get_user_field_update_internal_model(
    user_field: UserField = Depends(get_user_field_by_id_or_404),
):
    user_field_type = user_field.type
    configuration_type = UserFieldConfigurationBase
    if user_field_type == UserFieldType.CHOICE:
        configuration_type = UserFieldConfigurationChoice
    elif USER_FIELD_CAN_HAVE_DEFAULT[user_field_type]:
        configuration_type = UserFieldConfigurationDefault[
            USER_FIELD_TYPE_MAP[user_field_type]  # type: ignore
        ]

    return create_model(
        "UserFieldUpdateInternal",
        type=(Literal[user_field_type], ...),  # type: ignore
        configuration=(Optional[configuration_type], None),
        __base__=UserFieldUpdate,
    )


async def get_validated_user_field_update(
    user_field_update: UserFieldUpdate,
    user_field: UserField = Depends(get_user_field_by_id_or_404),
    user_field_update_internal_model: Type[UserFieldUpdate] = Depends(
        get_user_field_update_internal_model
    ),
) -> UserFieldUpdate:
    body_model = create_model(
        "UserFieldUpdateInternalBody",
        body=(user_field_update_internal_model, ...),
    )
    try:
        validated_user_field_update = body_model(
            body={"type": user_field.type, **user_field_update.dict(exclude_unset=True)}
        )
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_field_update.body  # type: ignore


async def get_registration_user_fields(
    manager: UserFieldManager = Depends(get_user_field_manager),
) -> List[UserField]:
    return await manager.get_registration_fields()


async def get_update_user_fields(
    manager: UserFieldManager = Depends(get_user_field_manager),
) -> List[UserField]:
    return await manager.get_update_fields()


def _get_pydantic_specification(user_fields: List[UserField]) -> Tuple[Any, Any]:
    fields: Any = {}
    validators: Any = {}
    for field in user_fields:
        field_type = get_user_field_pydantic_type(field)
        required = field.get_required()
        default = None

        if field.type == UserFieldType.BOOLEAN:
            """
            Special handling for booleans that are passed through HTML checkboxes.

            For the browser, when unchecked, the value is simply not passed.
            That's why we make it not required with a default at "False".

            When checked, the value is "on" by default, which is handled natively by Pydantic.

            Besides, we attach a special meaning to "required" booleans:
            it means that the value MUST BE "True".
            It can be useful for consent checkboxes for example.
            That's why in this case, we add a custom validator checking the value
            is only "True".
            """
            if required:
                validators[f"{field.slug}_validator"] = validator(
                    field.slug, allow_reuse=True, always=True
                )(true_bool_validator)
            required = False
            default = False

        fields[field.slug] = (
            field_type if required else Optional[field_type],
            default if default is not None else (... if required else None),
        )
    return fields, validators


async def get_user_create_model(
    registration_user_fields: List[UserField] = Depends(get_registration_user_fields),
) -> Type[UserCreate[UF]]:
    fields, validators = _get_pydantic_specification(registration_user_fields)
    user_fields_model = create_model(
        "UserFields",
        **fields,
        __validators__=validators,
        __base__=UserFields,  # type: ignore
    )
    return UserCreate[user_fields_model]  # type: ignore


async def get_user_create_internal_model(
    registration_user_fields: List[UserField] = Depends(get_registration_user_fields),
) -> Type[UserCreateInternal[UF]]:
    fields, validators = _get_pydantic_specification(registration_user_fields)
    user_fields_model = create_model(
        "UserFields",
        **fields,
        __validators__=validators,
        __base__=UserFields,  # type: ignore
    )
    return UserCreateInternal[user_fields_model]  # type: ignore


async def get_admin_user_create_internal_model(
    user_fields: List[UserField] = Depends(get_user_fields),
) -> Type[UserCreateInternal[UF]]:
    fields, validators = _get_pydantic_specification(user_fields)
    user_fields_model = create_model(
        "UserFields",
        **fields,
        __validators__=validators,
        __base__=UserFields,  # type: ignore
    )
    return UserCreateInternal[user_fields_model]  # type: ignore


async def get_user_update_model(
    update_user_fields: List[UserField] = Depends(get_update_user_fields),
) -> Type[UserUpdate[UF]]:
    fields, validators = _get_pydantic_specification(update_user_fields)
    user_fields_model = create_model(
        "UserFields",
        **fields,
        __validators__=validators,
        __base__=UserFields,  # type: ignore
    )
    return UserUpdate[user_fields_model]  # type: ignore


async def get_admin_user_update_model(
    user_fields: List[UserField] = Depends(get_user_fields),
) -> Type[UserUpdate[UF]]:
    fields, validators = _get_pydantic_specification(user_fields)
    user_fields_model = create_model(
        "UserFields",
        **fields,
        __validators__=validators,
        __base__=UserFields,  # type: ignore
    )
    return UserUpdate[user_fields_model]  # type: ignore



from typing import Any, Dict, Optional

from fastapi import Depends, HTTPException, status
from fastapi.security import APIKeyCookie

from fief.crypto.token import get_token_hash
from fief.dependencies.main_managers import get_admin_session_token_manager
from fief.managers import AdminSessionTokenManager
from fief.models import AdminSessionToken
from fief.settings import settings

cookie_scheme = APIKeyCookie(
    name=settings.fief_admin_session_cookie_name, auto_error=False
)


async def get_optional_admin_session_token(
    token: Optional[str] = Depends(cookie_scheme),
    manager: AdminSessionTokenManager = Depends(get_admin_session_token_manager),
) -> Optional[AdminSessionToken]:
    if token is None:
        return None
    token_hash = get_token_hash(token)
    session_token = await manager.get_by_token(token_hash)
    return session_token


async def get_admin_session_token(
    admin_session_token: Optional[AdminSessionToken] = Depends(
        get_optional_admin_session_token
    ),
) -> AdminSessionToken:
    if admin_session_token is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
    return admin_session_token


async def get_userinfo(
    session_token: AdminSessionToken = Depends(get_admin_session_token),
) -> Dict[str, Any]:
    return session_token.userinfo



from datetime import datetime, timezone
from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import SessionToken


class SessionTokenManager(BaseManager[SessionToken], UUIDManagerMixin[SessionToken]):
    model = SessionToken

    async def get_by_token(
        self, token: str, *, fresh: bool = True
    ) -> Optional[SessionToken]:
        statement = select(SessionToken).where(SessionToken.token == token)
        if fresh:
            statement = statement.where(
                SessionToken.expires_at > datetime.now(timezone.utc)
            )

        return await self.get_one_or_none(statement)



from datetime import datetime, timezone
from typing import List, Optional, Tuple

from fastapi import Cookie, Depends, Form, Query, Response

from fief.dependencies.authentication_flow import get_authentication_flow
from fief.dependencies.locale import get_gettext
from fief.dependencies.session_token import get_session_token
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.workspace_managers import (
    get_client_manager,
    get_grant_manager,
    get_login_session_manager,
)
from fief.exceptions import (
    AuthorizeException,
    AuthorizeRedirectException,
    ConsentException,
    LoginException,
)
from fief.managers import ClientManager, GrantManager, LoginSessionManager
from fief.models import Client, LoginSession, SessionToken, Tenant
from fief.schemas.auth import (
    AuthorizeError,
    AuthorizeRedirectError,
    ConsentError,
    LoginError,
)
from fief.services.authentication_flow import AuthenticationFlow
from fief.services.response_type import (
    ALLOWED_RESPONSE_TYPES,
    DEFAULT_RESPONSE_MODE,
    NONCE_REQUIRED_RESPONSE_TYPES,
)
from fief.settings import settings


async def get_authorize_client(
    client_id: Optional[str] = Query(None),
    manager: ClientManager = Depends(get_client_manager),
    _=Depends(get_gettext),
) -> Client:
    if client_id is None:
        raise AuthorizeException(
            AuthorizeError.get_invalid_client(_("client_id is missing"))
        )

    client = await manager.get_by_client_id(client_id)

    if client is None:
        raise AuthorizeException(AuthorizeError.get_invalid_client(_("Unknown client")))

    return client


async def get_authorize_redirect_uri(
    redirect_uri: Optional[str] = Query(None),
    client: Client = Depends(get_authorize_client),
    _=Depends(get_gettext),
) -> str:
    if redirect_uri is None:
        raise AuthorizeException(
            AuthorizeError.get_invalid_redirect_uri(_("redirect_uri is missing"))
        )

    if redirect_uri not in client.redirect_uris:
        raise AuthorizeException(
            AuthorizeError.get_invalid_redirect_uri(
                _("redirect_uri is not authorized for this client")
            )
        )

    return redirect_uri


async def get_authorize_state(state: Optional[str] = Query(None)) -> Optional[str]:
    return state


async def get_authorize_response_type(
    response_type: Optional[str] = Query(None),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> str:
    if response_type is None:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(_("response_type is missing")),
            redirect_uri,
            "query",
            state,
        )

    if response_type not in ALLOWED_RESPONSE_TYPES:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(_("response_type is invalid")),
            redirect_uri,
            "query",
            state,
        )

    return response_type


async def get_authorize_response_mode(
    response_type: str = Depends(get_authorize_response_type),
) -> str:
    return DEFAULT_RESPONSE_MODE[response_type]


async def check_unsupported_request_parameter(
    request_parameter: Optional[str] = Query(None, alias="request"),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> None:
    if request_parameter is not None:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_request_not_supported(
                _("request parameter is not supported")
            ),
            redirect_uri,
            response_mode,
            state,
        )


async def get_nonce(
    nonce: Optional[str] = Query(None),
    response_type: str = Depends(get_authorize_response_type),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> Optional[str]:
    if nonce is None and response_type in NONCE_REQUIRED_RESPONSE_TYPES:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(
                _("nonce parameter is required for this response_type")
            ),
            redirect_uri,
            response_mode,
            state,
        )

    return nonce


async def get_authorize_scope(
    scope: Optional[str] = Query(None),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> List[str]:
    if scope is None:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(_("scope is missing")),
            redirect_uri,
            response_mode,
            state,
        )

    scope_list = scope.split()

    if "openid" not in scope_list:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_scope(
                _('scope should contain "openid"')
            ),
            redirect_uri,
            response_mode,
            state,
        )

    return scope_list


async def get_authorize_prompt(
    prompt: Optional[str] = Query(None),
    session_token: Optional[SessionToken] = Depends(get_session_token),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> Optional[str]:
    if prompt is not None and prompt not in ["none", "login", "consent"]:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(
                _('prompt should either be "none", "login" or "register"')
            ),
            redirect_uri,
            response_mode,
            state,
        )

    if prompt in ["none", "consent"] and session_token is None:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_login_required(_("User is not logged in")),
            redirect_uri,
            response_mode,
            state,
        )

    return prompt


async def get_authorize_screen(
    screen: str = Query("login"),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> str:
    if screen not in ["login", "register"]:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(
                _('screen should either be "login" or "register"')
            ),
            redirect_uri,
            response_mode,
            state,
        )

    return screen


async def get_authorize_code_challenge(
    code_challenge: Optional[str] = Query(None),
    code_challenge_method: Optional[str] = Query("plain"),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    state: Optional[str] = Depends(get_authorize_state),
    _=Depends(get_gettext),
) -> Optional[Tuple[str, str]]:
    if code_challenge is None:
        return None

    if code_challenge_method not in ["plain", "S256"]:
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_invalid_request(
                _("Unsupported code_challenge_method")
            ),
            redirect_uri,
            response_mode,
            state,
        )

    return code_challenge, code_challenge_method


async def has_valid_session_token(
    max_age: Optional[int] = Query(None),
    session_token: Optional[SessionToken] = Depends(get_session_token),
) -> bool:
    if session_token is None:
        return False

    if max_age is not None:
        session_age = (
            datetime.now(timezone.utc) - session_token.created_at
        ).total_seconds()
        return session_age < max_age

    return True


async def get_optional_login_session(
    token: Optional[str] = Cookie(None, alias=settings.login_session_cookie_name),
    login_session_manager: LoginSessionManager = Depends(get_login_session_manager),
    tenant: Tenant = Depends(get_current_tenant),
) -> Optional[LoginSession]:
    if token is None:
        return None

    login_session = await login_session_manager.get_by_token(token)
    if login_session is None:
        return None

    if login_session.client.tenant_id != tenant.id:
        return None

    return login_session


async def get_login_session(
    login_session: Optional[LoginSession] = Depends(get_optional_login_session),
    _=Depends(get_gettext),
) -> LoginSession:
    if login_session is None:
        raise LoginException(
            LoginError.get_invalid_session(_("Invalid login session")), fatal=True
        )

    return login_session


async def get_consent_action(
    action: Optional[str] = Form(None),
    login_session: LoginSession = Depends(get_login_session),
    tenant: Tenant = Depends(get_current_tenant),
    _=Depends(get_gettext),
) -> str:
    if action is None or action not in ["allow", "deny"]:
        raise ConsentException(
            ConsentError.get_invalid_action(
                _('action should either be "allow" or "deny"')
            ),
            login_session.client,
            login_session.scope,
            tenant,
        )

    return action


async def get_needs_consent(
    login_session: LoginSession = Depends(get_login_session),
    session_token: Optional[SessionToken] = Depends(get_session_token),
    grant_manager: GrantManager = Depends(get_grant_manager),
) -> bool:
    if session_token is None:
        return True

    client = login_session.client

    if client.first_party:
        return False

    client_id = client.id
    user_id = session_token.user_id
    grant = await grant_manager.get_by_user_and_client(user_id, client_id)

    if grant is None or not set(login_session.scope).issubset(set(grant.scope)):
        return True

    return False


async def get_consent_prompt(
    login_session: LoginSession = Depends(get_login_session),
    needs_consent: bool = Depends(get_needs_consent),
    tenant: Tenant = Depends(get_current_tenant),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    _=Depends(get_gettext),
) -> Optional[str]:
    prompt = login_session.prompt
    if needs_consent and prompt == "none":
        await authentication_flow.delete_login_session(Response(), login_session)
        raise AuthorizeRedirectException(
            AuthorizeRedirectError.get_consent_required(
                _("User consent is required for this scope")
            ),
            login_session.redirect_uri,
            login_session.response_mode,
            login_session.state,
            tenant,
        )

    return prompt



import re

from fastapi.middleware.cors import CORSMiddleware
from starlette.types import Receive, Scope, Send


class CORSMiddlewarePath(CORSMiddleware):
    def __init__(self, path_regex: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.path_regex = re.compile(path_regex)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        path = scope["path"]
        if not self.path_regex.match(path):
            await self.app(scope, receive, send)
            return

        return await super().__call__(scope, receive, send)



from pydantic import UUID4
from sqlalchemy import Column, ForeignKey, String

from fief.models.base import MainBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.workspace import Workspace


class AdminAPIKey(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "admin_api_key"

    name: str = Column(String(length=255), nullable=False)
    token: str = Column(String(length=255), unique=True, nullable=False)
    workspace_id: UUID4 = Column(GUID, ForeignKey(Workspace.id, ondelete="CASCADE"), nullable=False)  # type: ignore

    def __repr__(self) -> str:
        return f"AdminAPIKey(id={self.id})"



from pathlib import Path

ROOT_DIR = Path(__file__).parent

ALEMBIC_CONFIG_FILE = str(ROOT_DIR / "alembic.ini")
STATIC_DIRECTORY = ROOT_DIR / "static"
LOCALE_DIRECTORY = ROOT_DIR / "locale"
TEMPLATES_DIRECTORY = ROOT_DIR / "templates"
EMAIL_TEMPLATES_DIRECTORY = ROOT_DIR / "emails"



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from fief.apps.admin.routers.api_keys import router as api_keys_router
from fief.apps.admin.routers.auth import router as auth_router
from fief.apps.admin.routers.clients import router as clients_router
from fief.apps.admin.routers.tenants import router as tenants_router
from fief.apps.admin.routers.user_fields import router as user_fields_router
from fief.apps.admin.routers.users import router as users_router
from fief.apps.admin.routers.workspaces import router as workspaces_router
from fief.settings import settings

app = FastAPI(title="Fief Administration API")

app.add_middleware(
    CORSMiddleware,
    allow_origin_regex=settings.allow_origin_regex,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_keys_router, prefix="/api-keys", include_in_schema=False)
app.include_router(auth_router, prefix="/auth", include_in_schema=False)
app.include_router(clients_router, prefix="/clients")
app.include_router(tenants_router, prefix="/tenants")
app.include_router(users_router, prefix="/users")
app.include_router(user_fields_router, prefix="/user-fields")
app.include_router(workspaces_router, prefix="/workspaces", include_in_schema=False)

__all__ = ["app"]



import uuid

import dramatiq

from fief.tasks.base import TaskBase


class OnAfterRegisterTask(TaskBase):
    __name__ = "on_after_register"

    async def run(self, user_id: str, workspace_id: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        user = await self._get_user(uuid.UUID(user_id), workspace)
        tenant = await self._get_tenant(user.tenant_id, workspace)

        translations = user.get_preferred_translations()
        title = translations.gettext("Welcome to %(tenant)s!") % {"tenant": tenant.name}
        context = {"tenant": tenant, "title": title}
        html = self._render_email_template("welcome.html", translations, context)
        self.email_provider.send_email(
            sender=("contact@fief.dev", None),
            recipient=(user.email, None),
            subject=title,
            html=html,
        )


on_after_register = dramatiq.actor(OnAfterRegisterTask())



from typing import Optional

from pydantic import BaseModel, root_validator, validator

from fief.crypto.encryption import decrypt
from fief.db.types import DatabaseType
from fief.errors import APIErrorCode
from fief.schemas.generics import UUIDSchema
from fief.settings import settings


def validate_all_database_settings(cls, values):
    database_type = values.get("database_type")
    database_settings = [
        values.get("database_host"),
        values.get("database_port"),
        values.get("database_username"),
        values.get("database_password"),
        values.get("database_name"),
    ]

    if database_type is None and not any(database_settings):
        return values

    if database_type is None and any(database_settings):
        raise ValueError(APIErrorCode.WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS)

    database_name = values.get("database_name")
    if database_type == DatabaseType.SQLITE:
        if database_name is None:
            raise ValueError(APIErrorCode.WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS)
    else:
        if not all(database_settings):
            raise ValueError(APIErrorCode.WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS)

    return values


class WorkspaceCheckConnection(BaseModel):
    database_type: DatabaseType
    database_host: str
    database_port: int
    database_username: str
    database_password: str
    database_name: str

    _validate_all_database_settings = root_validator(allow_reuse=True)(
        validate_all_database_settings
    )


class WorkspaceCreate(BaseModel):
    name: str
    database_type: Optional[DatabaseType]
    database_host: Optional[str]
    database_port: Optional[int]
    database_username: Optional[str]
    database_password: Optional[str]
    database_name: Optional[str]

    _validate_all_database_settings = root_validator(allow_reuse=True)(
        validate_all_database_settings
    )


class BaseWorkspace(UUIDSchema):
    name: str
    domain: str


class Workspace(BaseWorkspace):
    database_type: Optional[DatabaseType]
    database_host: Optional[str]
    database_port: Optional[int]
    database_username: Optional[str]
    database_password: Optional[str]
    database_name: Optional[str]

    @validator(
        "database_host",
        "database_username",
        "database_password",
        "database_name",
        pre=True,
    )
    def decrypt_database_setting(cls, value: Optional[str]) -> Optional[str]:
        if value is None:
            return value
        return decrypt(value, settings.encryption_key)

    @validator("database_port", pre=True)
    def decrypt_database_port(cls, value: Optional[str]) -> Optional[int]:
        if value is None:
            return value
        return int(decrypt(value, settings.encryption_key))


class WorkspacePublic(BaseWorkspace):
    pass



import uuid
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional, TypeVar

from pydantic import UUID4
from sqlalchemy import TIMESTAMP, Column
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import declarative_mixin
from sqlalchemy.sql import func
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(36), storing as regular strings.
    """

    class UUIDChar(CHAR):
        python_type = UUID4  # type: ignore

    impl = UUIDChar
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(36))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return str(uuid.UUID(value))
            else:
                return str(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class BaseModel:
    pass


@declarative_mixin
class UUIDModel(BaseModel):
    if TYPE_CHECKING:
        id: UUID4
    else:
        id: UUID4 = Column(GUID, primary_key=True, default=uuid.uuid4)


def now_utc():
    return datetime.now(timezone.utc)


class TIMESTAMPAware(TypeDecorator):
    """
    MySQL and SQLite will always return naive-Python datetimes.

    We store everything as UTC, but we want to have
    only offset-aware Python datetimes, even with MySQL and SQLite.
    """

    impl = TIMESTAMP
    cache_ok = True

    def process_result_value(self, value: Optional[datetime], dialect):
        if value is not None and dialect.name != "postgresql":
            return value.replace(tzinfo=timezone.utc)
        return value


@declarative_mixin
class CreatedUpdatedAt(BaseModel):
    if TYPE_CHECKING:
        created_at: datetime
        updated_at: datetime
    else:
        created_at: datetime = Column(
            TIMESTAMPAware(timezone=True),
            nullable=False,
            index=True,
            default=now_utc,
            server_default=func.now(),
        )
        updated_at: datetime = Column(
            TIMESTAMPAware(timezone=True),
            nullable=False,
            index=True,
            default=now_utc,
            server_default=func.now(),
            onupdate=now_utc,
        )


M = TypeVar("M", bound=BaseModel)
M_UUID = TypeVar("M_UUID", bound=UUIDModel)



import re
from typing import List, Optional

from pydantic import UUID4, AnyUrl, BaseModel, Field, SecretStr, validator

from fief.errors import APIErrorCode
from fief.models.client import ClientType
from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema
from fief.schemas.tenant import TenantEmbedded

LOCALHOST_HOST_PATTERN = re.compile(r"([^\.]+\.)?localhost(\d+)?", flags=re.IGNORECASE)


def validate_redirect_uri(url: AnyUrl) -> AnyUrl:
    if url.scheme == "http" and (
        url.host is None or not LOCALHOST_HOST_PATTERN.match(url.host)
    ):
        raise ValueError(APIErrorCode.CLIENT_HTTPS_REQUIRED_ON_REDIRECT_URIS.value)
    return url


class ClientCreate(BaseModel):
    name: str
    first_party: bool
    client_type: ClientType
    redirect_uris: List[AnyUrl] = Field(..., min_items=1)
    tenant_id: UUID4

    _validate_redirect_uri = validator(
        "redirect_uris", each_item=True, allow_reuse=True
    )(validate_redirect_uri)


class ClientUpdate(BaseModel):
    name: Optional[str]
    first_party: Optional[bool]
    client_type: Optional[ClientType]
    redirect_uris: Optional[List[AnyUrl]] = Field(None, min_items=1)

    _validate_redirect_uri = validator(
        "redirect_uris", each_item=True, allow_reuse=True
    )(validate_redirect_uri)


class BaseClient(UUIDSchema, CreatedUpdatedAt):
    name: str
    first_party: bool
    client_type: ClientType
    client_id: str
    client_secret: str
    redirect_uris: List[AnyUrl]
    tenant_id: UUID4


class Client(BaseClient):
    tenant: TenantEmbedded
    encrypt_jwk: Optional[SecretStr]



from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema


class BaseTenant(UUIDSchema, CreatedUpdatedAt):
    name: str
    default: bool
    slug: str


class Tenant(BaseTenant):
    pass


class TenantEmbedded(BaseTenant):
    pass



import json

from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse

from fief.crypto.token import generate_token
from fief.dependencies.admin_session import get_admin_session_token, get_userinfo
from fief.dependencies.fief import FiefAsyncRelativeEndpoints, get_fief
from fief.dependencies.main_managers import get_admin_session_token_manager
from fief.managers import AdminSessionTokenManager
from fief.models import AdminSessionToken
from fief.settings import settings

router = APIRouter()


@router.get("/login", name="admin.auth:login")
async def login(
    request: Request,
    screen: str = Query("login"),
    fief: FiefAsyncRelativeEndpoints = Depends(get_fief),
):
    url = await fief.auth_url(
        redirect_uri=request.url_for("admin.auth:callback"),
        scope=["openid"],
        extras_params={"screen": screen},
    )
    return RedirectResponse(url=url, status_code=status.HTTP_302_FOUND)


@router.get("/callback", name="admin.auth:callback")
async def callback(
    request: Request,
    code: str = Query(...),
    fief: FiefAsyncRelativeEndpoints = Depends(get_fief),
    manager: AdminSessionTokenManager = Depends(get_admin_session_token_manager),
):
    tokens, userinfo = await fief.auth_callback(
        code, request.url_for("admin.auth:callback")
    )
    token, token_hash = generate_token()
    session_token = AdminSessionToken(
        token=token_hash,
        raw_tokens=json.dumps(tokens),
        raw_userinfo=json.dumps(userinfo),
    )
    await manager.create(session_token)

    response = RedirectResponse(url="/admin/", status_code=status.HTTP_302_FOUND)
    response.set_cookie(
        settings.fief_admin_session_cookie_name,
        token,
        domain=settings.fief_admin_session_cookie_domain,
        secure=settings.fief_admin_session_cookie_secure,
        httponly=True,
    )

    return response


@router.get("/userinfo", name="admin.auth:userinfo")
async def userinfo(userinfo=Depends(get_userinfo)):
    return userinfo


@router.get("/logout", name="admin.auth:logout")
async def logout(
    request: Request,
    session_token: AdminSessionToken = Depends(get_admin_session_token),
    manager: AdminSessionTokenManager = Depends(get_admin_session_token_manager),
):
    await manager.delete(session_token)

    response = RedirectResponse(
        url=f"//{settings.fief_domain}/logout?redirect_uri={request.base_url}admin/",
        status_code=status.HTTP_302_FOUND,
    )
    response.delete_cookie(
        settings.fief_admin_session_cookie_name,
        domain=settings.fief_admin_session_cookie_domain,
        secure=settings.fief_admin_session_cookie_secure,
        httponly=True,
    )

    return response



import random
import string
from typing import Optional

from slugify import slugify
from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import Workspace
from fief.settings import settings


class WorkspaceManager(BaseManager[Workspace], UUIDManagerMixin[Workspace]):
    model = Workspace

    async def get_by_domain(self, domain: str) -> Optional[Workspace]:
        statement = select(Workspace).where(Workspace.domain == domain)
        return await self.get_one_or_none(statement)

    async def get_main(self) -> Optional[Workspace]:
        return await self.get_by_domain(settings.fief_domain)

    async def get_available_subdomain(self, name: str) -> str:
        slug = slugify(name)
        domain = f"{slug}.{settings.root_domain}"
        workspace = await self.get_by_domain(domain)

        if workspace is None:
            return domain

        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=6)
        )
        return f"{slug}-{random_string}.{settings.root_domain}"



from typing import Any, Dict, List, Type

from fastapi import Depends
from pydantic import ValidationError

from fief.dependencies.form import get_form_data_dict
from fief.dependencies.locale import get_translations
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.user_field import (
    get_registration_user_fields,
    get_user_create_model,
)
from fief.exceptions import FormValidationError
from fief.locale import Translations
from fief.models import Tenant, UserField
from fief.schemas.user import UF, UserCreate


async def get_register_context(
    registration_user_fields: List[UserField] = Depends(get_registration_user_fields),
    tenant: Tenant = Depends(get_current_tenant),
) -> Dict[str, Any]:
    return {
        "tenant": tenant,
        "registration_user_fields": registration_user_fields,
    }


async def get_user_create(
    form_data_dict: Dict[str, Any] = Depends(get_form_data_dict),
    user_create_model: Type[UserCreate[UF]] = Depends(get_user_create_model),
    register_context: Dict[str, Any] = Depends(get_register_context),
    translations: Translations = Depends(get_translations),
) -> UserCreate[UF]:
    try:
        return user_create_model(**form_data_dict)
    except ValidationError as e:
        raise FormValidationError(
            "register.html",
            {
                "form_data": form_data_dict,
                **register_context,
            },
            translations,
            e.raw_errors,
            user_create_model,
        ) from e



import asyncio
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from pydantic import UUID4
from sqlalchemy import func, select
from sqlalchemy.engine import Result
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute, RelationshipProperty
from sqlalchemy.sql import Select

from fief.models.generics import M_UUID, M


class BaseManagerProtocol(Protocol[M]):
    model: Type[M]
    session: AsyncSession

    async def paginate(
        self,
        statement: Select,
        limit=10,
        skip=0,
    ) -> Tuple[List[M], int]:
        ...  # pragma: no cover

    def orderize(
        self, statement: Select, ordering: List[Tuple[List[str], bool]]
    ) -> Select:
        ...  # pragma: no cover

    async def get_one_or_none(self, statement: Select) -> Optional[M]:
        ...  # pragma: no cover

    async def list(self, statement: Select) -> List[M]:
        ...  # pragma: no cover

    async def create(self, object: M) -> M:
        ...  # pragma: no cover

    async def update(self, object: M) -> None:
        ...  # pragma: no cover

    async def delete(self, object: M) -> None:
        ...  # pragma: no cover

    async def _execute_statement(self, statement: Select) -> Result:
        ...  # pragma: no cover


class UUIDManagerProtocol(BaseManagerProtocol, Protocol[M_UUID]):
    model: Type[M_UUID]

    async def get_by_id(self, id: UUID4) -> Optional[M_UUID]:
        ...  # pragma: no cover


class BaseManager(BaseManagerProtocol, Generic[M]):
    def __init__(self, session: AsyncSession) -> None:
        self.session = session

    async def paginate(
        self,
        statement: Select,
        limit=10,
        skip=0,
    ) -> Tuple[List[M], int]:
        paginated_statement = statement.offset(skip).limit(limit)
        [count, results] = await asyncio.gather(
            self._count(statement), self._execute_statement(paginated_statement)
        )

        return [result[0] for result in results.unique().all()], count

    def orderize(
        self, statement: Select, ordering: List[Tuple[List[str], bool]]
    ) -> Select:
        for (accessors, is_desc) in ordering:
            field: InstrumentedAttribute
            # Local field
            if len(accessors) == 1:
                try:
                    field = getattr(self.model, accessors[0])
                    if not isinstance(
                        field.prop, RelationshipProperty
                    ):  # Prevent ordering by raw relation ship field -> "tenant" instead of "tenant.id"
                        statement = statement.order_by(
                            field.desc() if is_desc else field.asc()
                        )
                except AttributeError:
                    pass
            # Relationship field
            else:
                valid_field = True
                model = self.model
                for accessor in accessors:
                    try:
                        field = getattr(model, accessor)
                        if isinstance(field.prop, RelationshipProperty):
                            if field.prop.lazy != "joined":
                                statement = statement.join(field)
                            model = field.prop.entity.class_
                    except AttributeError:
                        valid_field = False
                        break
                if valid_field:
                    statement = statement.order_by(
                        field.desc() if is_desc else field.asc()
                    )
        return statement

    async def all(self) -> List[M]:
        return await self.list(select(self.model))

    async def get_one_or_none(self, statement: Select) -> Optional[M]:
        results = await self._execute_statement(statement)
        object = results.first()
        if object is None:
            return None
        return object[0]

    async def list(self, statement: Select) -> List[M]:
        results = await self._execute_statement(statement)
        return [result[0] for result in results.unique().all()]

    async def create(self, object: M) -> M:
        self.session.add(object)
        await self.session.commit()
        await self.session.refresh(object)
        return object

    async def update(self, object: M) -> None:
        self.session.add(object)
        await self.session.commit()
        await self.session.refresh(object)

    async def delete(self, object: M) -> None:
        await self.session.delete(object)
        await self.session.commit()

    async def _count(self, statement: Select) -> int:
        count_statement = statement.with_only_columns(
            [func.count()], maintain_column_froms=True  # type: ignore
        ).order_by(None)
        results = await self._execute_statement(count_statement)
        return results.scalar_one()

    async def _execute_statement(self, statement: Select) -> Result:
        return await self.session.execute(statement)


class UUIDManagerMixin(Generic[M_UUID]):
    async def get_by_id(
        self: UUIDManagerProtocol[M_UUID],
        id: UUID4,
        options: Optional[Sequence[Any]] = None,
    ) -> Optional[M_UUID]:
        statement = select(self.model).where(self.model.id == id)

        if options is not None:
            statement = statement.options(*options)

        return await self.get_one_or_none(statement)


MANAGER = TypeVar("MANAGER", bound=BaseManager)


def get_manager(manager_class: Type[MANAGER], session: AsyncSession) -> MANAGER:
    return manager_class(session)



from datetime import datetime, timedelta, timezone
from typing import List, Optional, Tuple, TypeVar

from fastapi import Response, status
from fastapi.responses import RedirectResponse
from furl import furl
from pydantic import UUID4

from fief.crypto.access_token import generate_access_token
from fief.crypto.id_token import generate_id_token, get_validation_hash
from fief.crypto.token import generate_token
from fief.managers import (
    AuthorizationCodeManager,
    GrantManager,
    LoginSessionManager,
    SessionTokenManager,
)
from fief.models import (
    AuthorizationCode,
    Client,
    Grant,
    LoginSession,
    SessionToken,
    Tenant,
    User,
    Workspace,
)
from fief.settings import settings

ResponseType = TypeVar("ResponseType", bound=Response)


class AuthenticationFlow:
    def __init__(
        self,
        authorization_code_manager: AuthorizationCodeManager,
        login_session_manager: LoginSessionManager,
        session_token_manager: SessionTokenManager,
        grant_manager: GrantManager,
    ) -> None:
        self.authorization_code_manager = authorization_code_manager
        self.login_session_manager = login_session_manager
        self.session_token_manager = session_token_manager
        self.grant_manager = grant_manager

    async def create_login_session(
        self,
        response: ResponseType,
        *,
        response_type: str,
        response_mode: str,
        redirect_uri: str,
        scope: List[str],
        state: Optional[str],
        nonce: Optional[str],
        code_challenge_tuple: Optional[Tuple[str, str]],
        client: Client,
    ) -> ResponseType:
        login_session = LoginSession(
            response_type=response_type,
            response_mode=response_mode,
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
            nonce=nonce,
            client_id=client.id,
        )
        if code_challenge_tuple is not None:
            code_challenge, code_challenge_method = code_challenge_tuple
            login_session.code_challenge = code_challenge
            login_session.code_challenge_method = code_challenge_method

        login_session = await self.login_session_manager.create(login_session)

        response.set_cookie(
            settings.login_session_cookie_name,
            login_session.token,
            domain=settings.login_session_cookie_domain,
            secure=settings.login_session_cookie_secure,
        )

        return response

    async def delete_login_session(
        self, response: ResponseType, login_session: LoginSession
    ) -> ResponseType:
        response.delete_cookie(
            settings.login_session_cookie_name,
            domain=settings.login_session_cookie_domain,
        )
        await self.login_session_manager.delete(login_session)
        return response

    async def create_session_token(
        self, response: ResponseType, user_id: UUID4
    ) -> ResponseType:
        token, token_hash = generate_token()
        expires_at = datetime.now(timezone.utc) + timedelta(
            seconds=settings.session_lifetime_seconds
        )
        await self.session_token_manager.create(
            SessionToken(token=token_hash, expires_at=expires_at, user_id=user_id)
        )
        response.set_cookie(
            settings.session_cookie_name,
            value=token,
            max_age=settings.session_lifetime_seconds,
            domain=settings.session_cookie_domain,
            secure=settings.session_cookie_secure,
            httponly=True,
        )

        return response

    async def rotate_session_token(
        self,
        response: ResponseType,
        user_id: UUID4,
        *,
        session_token: Optional[SessionToken],
    ) -> ResponseType:
        if session_token is not None:
            await self.session_token_manager.delete(session_token)
        return await self.create_session_token(response, user_id)

    async def create_or_update_grant(
        self, user_id: UUID4, client: Client, scope: List[str]
    ) -> Grant:
        grant = await self.grant_manager.get_by_user_and_client(user_id, client.id)
        if grant is not None:
            grant.scope = scope
            await self.grant_manager.update(grant)
        else:
            grant = await self.grant_manager.create(
                Grant(scope=scope, user_id=user_id, client=client)
            )
        return grant

    async def get_authorization_code_success_redirect(
        self,
        *,
        login_session: LoginSession,
        authenticated_at: datetime,
        user: User,
        client: Client,
        tenant: Tenant,
        workspace: Workspace,
    ) -> RedirectResponse:
        code, code_hash = generate_token()
        c_hash = get_validation_hash(code)
        authorization_code = await self.authorization_code_manager.create(
            AuthorizationCode(
                code=code_hash,
                c_hash=c_hash,
                redirect_uri=login_session.redirect_uri,
                scope=login_session.scope,
                authenticated_at=authenticated_at,
                nonce=login_session.nonce,
                user_id=user.id,
                client_id=client.id,
            )
        )

        code_challenge_tuple = login_session.get_code_challenge_tuple()
        if code_challenge_tuple is not None:
            code_challenge, code_challenge_method = code_challenge_tuple
            authorization_code.code_challenge = code_challenge
            authorization_code.code_challenge_method = code_challenge_method

        params = {"code": code}

        if login_session.state is not None:
            params["state"] = login_session.state

        tenant_host = tenant.get_host(workspace.domain)
        access_token: Optional[str] = None

        if login_session.response_type in ["code token", "code id_token token"]:
            access_token = generate_access_token(
                tenant.get_sign_jwk(),
                tenant_host,
                client,
                user,
                login_session.scope,
                settings.access_id_token_lifetime_seconds,
            )
            params["access_token"] = access_token
            params["token_type"] = "bearer"

        if login_session.response_type in ["code id_token", "code id_token token"]:
            id_token = generate_id_token(
                tenant.get_sign_jwk(),
                tenant_host,
                client,
                authenticated_at,
                user,
                settings.access_id_token_lifetime_seconds,
                nonce=login_session.nonce,
                c_hash=c_hash,
                access_token=access_token,
                encryption_key=client.get_encrypt_jwk(),
            )
            params["id_token"] = id_token

        parsed_redirect_uri = furl(login_session.redirect_uri)

        if login_session.response_mode == "query":
            parsed_redirect_uri.query.add(params)
        elif login_session.response_mode == "fragment":
            parsed_redirect_uri.fragment.add(args=params)

        return RedirectResponse(
            url=parsed_redirect_uri.url, status_code=status.HTTP_302_FOUND
        )

    @classmethod
    def get_authorization_code_error_redirect(
        cls,
        redirect_uri: str,
        response_mode: str,
        error: str,
        *,
        error_description: Optional[str] = None,
        error_uri: Optional[str] = None,
        state: Optional[str] = None,
    ) -> RedirectResponse:
        parsed_redirect_uri = furl(redirect_uri)

        params = {"error": error}
        if error_description is not None:
            params["error_description"] = error_description
        if error_uri is not None:
            params["error_uri"] = error_uri
        if state is not None:
            params["state"] = state

        if response_mode == "query":
            parsed_redirect_uri.query.add(params)
        elif response_mode == "fragment":
            parsed_redirect_uri.fragment.add(args=params)

        return RedirectResponse(
            url=parsed_redirect_uri.url, status_code=status.HTTP_302_FOUND
        )



from typing import Optional

from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from pydantic import UUID4
from sqlalchemy import select

from fief.crypto.token import get_token_hash
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.main_managers import get_admin_api_key_manager
from fief.dependencies.pagination import (
    Ordering,
    PaginatedObjects,
    Pagination,
    get_ordering,
    get_paginated_objects,
    get_pagination,
)
from fief.managers import AdminAPIKeyManager
from fief.models import AdminAPIKey, Workspace

bearer_scheme = HTTPBearer(auto_error=False)


async def get_optional_admin_api_key(
    authorization: Optional[HTTPAuthorizationCredentials] = Depends(bearer_scheme),
    manager: AdminAPIKeyManager = Depends(get_admin_api_key_manager),
) -> Optional[AdminAPIKey]:
    if authorization is None:
        return None
    token_hash = get_token_hash(authorization.credentials)
    admin_api_key = await manager.get_by_token(token_hash)
    return admin_api_key


async def get_paginated_api_keys(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(get_ordering),
    manager: AdminAPIKeyManager = Depends(get_admin_api_key_manager),
    current_workspace: Workspace = Depends(get_current_workspace),
) -> PaginatedObjects[AdminAPIKey]:
    statement = select(AdminAPIKey).where(
        AdminAPIKey.workspace_id == current_workspace.id
    )
    return await get_paginated_objects(statement, pagination, ordering, manager)


async def get_api_key_by_id_or_404(
    id: UUID4,
    manager: AdminAPIKeyManager = Depends(get_admin_api_key_manager),
    current_workspace: Workspace = Depends(get_current_workspace),
) -> AdminAPIKey:
    statement = select(AdminAPIKey).where(
        AdminAPIKey.id == id,
        AdminAPIKey.workspace_id == current_workspace.id,
    )
    api_key = await manager.get_one_or_none(statement)

    if api_key is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return api_key



from fief.schemas import (
    admin_api_key,
    auth,
    client,
    generics,
    reset,
    tenant,
    user,
    user_field,
    well_known,
    workspace,
)

__all__ = [
    "workspace",
    "admin_api_key",
    "auth",
    "client",
    "generics",
    "reset",
    "tenant",
    "user",
    "user_field",
    "well_known",
]



from enum import Enum


class APIErrorCode(str, Enum):
    CSRF_CHECK_FAILED = "CSRF_CHECK_FAILED"
    CANT_DETERMINE_VALID_WORKSPACE = "CANT_DETERMINE_VALID_WORKSPACE"

    WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS = (
        "WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS"
    )
    WORKSPACE_DB_CONNECTION_ERROR = "WORKSPACE_DB_CONNECTION_ERROR"
    WORKSPACE_DB_OUTDATED_MIGRATION = "WORKSPACE_DB_OUTDATED_MIGRATION"

    CLIENT_CREATE_UNKNOWN_TENANT = "CLIENT_CREATE_UNKNOWN_TENANT"
    CLIENT_HTTPS_REQUIRED_ON_REDIRECT_URIS = "CLIENT_HTTPS_REQUIRED_ON_REDIRECT_URIS"

    USER_FIELD_SLUG_ALREADY_EXISTS = "USER_FIELD_SLUG_ALREADY_EXISTS"

    USER_CREATE_UNKNOWN_TENANT = "USER_CREATE_UNKNOWN_TENANT"
    USER_CREATE_ALREADY_EXISTS = "USER_CREATE_ALREADY_EXISTS"
    USER_CREATE_INVALID_PASSWORD = "USER_CREATE_INVALID_PASSWORD"

    USER_UPDATE_INVALID_PASSWORD = "USER_UPDATE_INVALID_PASSWORD"
    USER_UPDATE_EMAIL_ALREADY_EXISTS = "USER_UPDATE_EMAIL_ALREADY_EXISTS"
